@@ -45,6 +45,14 @@ class GitlabCLI(object):
         self.mgr_cls._path = self.mgr_cls._path % self.args
         self.mgr = self.mgr_cls(gl)
 
+        types = getattr(self.mgr_cls, '_types', {})
+        if types:
+            for attr_name, type_cls in types.items():
+                if attr_name in self.args.keys():
+                    obj = type_cls()
+                    obj.set_from_cli(self.args[attr_name])
+                    self.args[attr_name] = obj.get()
+
     def __call__(self):
         method = 'do_%s' % self.action
         if hasattr(self, method):
@@ -135,8 +143,9 @@ def _populate_sub_parser_by_class(cls, sub_parser):
                                            action='store_true')
 
         if action_name == 'delete':
-            id_attr = cls._id_attr.replace('_', '-')
-            sub_parser_action.add_argument("--%s" % id_attr, required=True)
+            if cls._id_attr is not None:
+                id_attr = cls._id_attr.replace('_', '-')
+                sub_parser_action.add_argument("--%s" % id_attr, required=True)
 
         if action_name == "get":
             if gitlab.mixins.GetWithoutIdMixin not in inspect.getmro(cls):
@@ -154,11 +163,11 @@ def _populate_sub_parser_by_class(cls, sub_parser):
             if hasattr(mgr_cls, '_create_attrs'):
                 [sub_parser_action.add_argument("--%s" % x.replace('_', '-'),
                                                 required=True)
-                 for x in mgr_cls._create_attrs[0] if x != cls._id_attr]
+                 for x in mgr_cls._create_attrs[0]]
 
                 [sub_parser_action.add_argument("--%s" % x.replace('_', '-'),
                                                 required=False)
-                 for x in mgr_cls._create_attrs[1] if x != cls._id_attr]
+                 for x in mgr_cls._create_attrs[1]]
 
         if action_name == "update":
             if cls._id_attr is not None:
@@ -241,6 +250,7 @@ def extend_parser(parser):
         object_group = subparsers.add_parser(arg_name)
 
         object_subparsers = object_group.add_subparsers(
+            title='action',
             dest='action', help="Action to execute.")
         _populate_sub_parser_by_class(cls, object_subparsers)
         object_subparsers.required = True
@@ -248,19 +258,37 @@ def extend_parser(parser):
     return parser
 
 
+def get_dict(obj, fields):
+    if isinstance(obj, six.string_types):
+        return obj
+
+    if fields:
+        return {k: v for k, v in obj.attributes.items()
+                if k in fields}
+    return obj.attributes
+
+
 class JSONPrinter(object):
     def display(self, d, **kwargs):
         import json  # noqa
-
         print(json.dumps(d))
 
+    def display_list(self, data, fields, **kwargs):
+        import json  # noqa
+        print(json.dumps([get_dict(obj, fields) for obj in data]))
+
 
 class YAMLPrinter(object):
     def display(self, d, **kwargs):
         import yaml  # noqa
-
         print(yaml.safe_dump(d, default_flow_style=False))
 
+    def display_list(self, data, fields, **kwargs):
+        import yaml  # noqa
+        print(yaml.safe_dump(
+            [get_dict(obj, fields) for obj in data],
+            default_flow_style=False))
+
 
 class LegacyPrinter(object):
     def display(self, d, **kwargs):
@@ -298,7 +326,23 @@ class LegacyPrinter(object):
                 print('%s: %s' % (obj._id_attr.replace('_', '-'), id))
             if hasattr(obj, '_short_print_attr'):
                 value = getattr(obj, obj._short_print_attr)
-                print('%s: %s' % (obj._short_print_attr, value))
+                value = value.replace('\r', '').replace('\n', ' ')
+                # If the attribute is a note (ProjectCommitComment) then we do
+                # some modifications to fit everything on one line
+                line = '%s: %s' % (obj._short_print_attr, value)
+                # ellipsize long lines (comments)
+                if len(line) > 79:
+                    line = line[:76] + '...'
+                print(line)
+
+    def display_list(self, data, fields, **kwargs):
+        verbose = kwargs.get('verbose', False)
+        for obj in data:
+            if isinstance(obj, gitlab.base.RESTObject):
+                self.display(get_dict(obj, fields), verbose=verbose, obj=obj)
+            else:
+                print(obj)
+            print('')
 
 
 PRINTERS = {
@@ -310,28 +354,15 @@ PRINTERS = {
 
 def run(gl, what, action, args, verbose, output, fields):
     g_cli = GitlabCLI(gl, what, action, args)
-    ret_val = g_cli()
+    data = g_cli()
 
     printer = PRINTERS[output]()
 
-    def get_dict(obj):
-        if fields:
-            return {k: v for k, v in obj.attributes.items()
-                    if k in fields}
-        return obj.attributes
-
-    if isinstance(ret_val, dict):
-        printer.display(ret_val, verbose=True, obj=ret_val)
-    elif isinstance(ret_val, list):
-        for obj in ret_val:
-            if isinstance(obj, gitlab.base.RESTObject):
-                printer.display(get_dict(obj), verbose=verbose, obj=obj)
-            else:
-                print(obj)
-            print('')
-    elif isinstance(ret_val, dict):
-        printer.display(ret_val, verbose=verbose, obj=ret_val)
-    elif isinstance(ret_val, gitlab.base.RESTObject):
-        printer.display(get_dict(ret_val), verbose=verbose, obj=ret_val)
-    elif isinstance(ret_val, six.string_types):
-        print(ret_val)
+    if isinstance(data, dict):
+        printer.display(data, verbose=True, obj=data)
+    elif isinstance(data, list):
+        printer.display_list(data, fields, verbose=verbose)
+    elif isinstance(data, gitlab.base.RESTObject):
+        printer.display(get_dict(data, fields), verbose=verbose, obj=data)
+    elif isinstance(data, six.string_types):
+        print(data)
